"""Whole-slide image streamer for TensorFlow."""

__version__ = "2.1.1"

"""

This module supports efficient whole-slide reading and processing in a tensorflow graph.

"""
from . import configure, tensorflow, codecs
